import numpy as np

NUM_PE = 1024
HEADER = "0001"
CHIP_X = "00001"
CHIP_Y = "00001"
NO_COPY = "00000"
DEF_POS = "00000"
INI_REF_PATH_COUNT = [96,64,32,0] # N,S,W,E
directions = ["N","S","W","E"]

def get_index_path(path_x : str, path_y : str):
    index_path = NUM_PE-1
    for i in range(len(path_x)):
        index_path -= (4**i)*(int(path_x[4-i])*2 + int(path_y[4-i]))
    return index_path

def get_ext_index_path(direction, dir_ref_count):
    path = dir_ref_count[direction]
    dir_ref_count[direction] += 1
    return path, dir_ref_count

def get_multicast_reference(copy_x, copy_y, recv_path, packet):
    x_paths = []
    y_paths = []
    x_poss_paths = []
    y_poss_paths = []
    num_paths = 1
    for bit_pos in range(len(copy_x)):
        if copy_x[bit_pos] == "1" and copy_y[bit_pos] == "1":
            x_poss_paths.append(["0","0","1","1"])
            y_poss_paths.append(["0","1","0","1"])
            num_paths *= 4
        elif copy_x[bit_pos] == "1":
            x_poss_paths.append(["0","1"])
            y_poss_paths.append([recv_path[1][bit_pos],recv_path[1][bit_pos]])
            num_paths *= 2
        elif copy_y[bit_pos] == "1":
            x_poss_paths.append([recv_path[0][bit_pos],recv_path[0][bit_pos]])
            y_poss_paths.append(["0","1"])
            num_paths *= 2
        else:
            x_poss_paths.append([recv_path[0][bit_pos]])
            y_poss_paths.append([recv_path[1][bit_pos]])

    division = num_paths
    for i in range(len(x_poss_paths)):
        x_poss_ind = 0
        y_poss_ind = 0
        division = int(division/len(x_poss_paths[i]))
        for j in range(num_paths):
            if i == 0:
                x_paths.append(x_poss_paths[0][x_poss_ind])
                y_paths.append(y_poss_paths[0][y_poss_ind])
            else:
                x_paths[j] += x_poss_paths[i][x_poss_ind]
                y_paths[j] += y_poss_paths[i][y_poss_ind]
            if ((j+1) % division) == 0:
                x_poss_ind  = x_poss_ind+1 if x_poss_ind+1 < len(x_poss_paths[i]) \
                            else 0
                y_poss_ind  = y_poss_ind+1 if y_poss_ind+1 < len(y_poss_paths[i]) \
                            else 0
    reference = ""
    for x_path, y_path in zip(x_paths, y_paths):
        index_path = get_index_path(x_path, y_path)
        reference += "0 " +  str(index_path) + " "
        reference += packet + "\n"
        
    return reference

def get_origin_dest_comment(path, recv_path):
    if len(path) == 1:
        comment = " # origin: (" + path + ")"
    else:
        comment = " # origin: (" + path[0] + "-" + path[1] + ")"
    
    comment += " --> dest: ("
    if len(recv_path) == 1:
        comment += recv_path
    else:
        for i in range(0, len(recv_path), 2):
            comment += recv_path[i] + "-" + recv_path[i+1]
            if i+2 < len(recv_path):
                comment += ","
    comment += ")"
    return comment

def gen_ref_for_ext_msg(path, recv_path, data_packet, stimuli, reference):
    if recv_path[0] == "N":
        ex_chip_x = CHIP_X
        ex_chip_y = f'{int(CHIP_Y,2)-1:05b}'
        index_path = INI_REF_PATH_COUNT[0]
    elif recv_path[0] == "S":
        ex_chip_x = CHIP_X
        ex_chip_y = f'{int(CHIP_Y,2)+1:05b}'
        index_path = INI_REF_PATH_COUNT[1]
    elif recv_path[0] == "W":
        ex_chip_x = f'{int(CHIP_X,2)-1:05b}'
        ex_chip_y = CHIP_Y
        index_path = INI_REF_PATH_COUNT[2]
    else:
        ex_chip_x = f'{int(CHIP_X,2)+1:05b}'
        ex_chip_y = CHIP_Y
        index_path = INI_REF_PATH_COUNT[3]
    packet = HEADER + ex_chip_x + ex_chip_y
    packet += NO_COPY*4
    # assign packet
    packet += f'{data_packet:030b}'
    stimuli += packet + get_origin_dest_comment(path, recv_path) + "\n"
    # reference
    reference += "1 " +  str(index_path) + " "
    reference += packet + "\n"

    return data_packet, stimuli, reference


def gen_ref_for_sing_dir(path, recv_path, data_packet, stimuli, reference):
    packet = HEADER + CHIP_X + CHIP_Y
    packet += recv_path[0] + recv_path[1] + NO_COPY*2
    # assign packet
    packet += f'{data_packet:030b}'
    stimuli += packet + get_origin_dest_comment(path, recv_path) + "\n"
    # reference       
    index_path = get_index_path(recv_path[0], recv_path[1])
    reference += "0 " +  str(index_path) + " "
    reference += packet + "\n"

    return data_packet, stimuli, reference


def gen_ref_for_mult_dir_multicast(path, recv_path, data_packet, stimuli, reference):
    list_copy_x = list("00000")
    list_copy_y = list("00000")
    for j in range(2,len(recv_path),2):
        for i in range(len(recv_path[0])):
            if recv_path[0][i] != recv_path[j][i]:
                list_copy_x[i]= "1"
            if recv_path[1][i] != recv_path[j+1][i]:
                list_copy_y[i]= "1"

    copy_x = ''.join(list_copy_x)
    copy_y = ''.join(list_copy_y)
    packet = HEADER + CHIP_X + CHIP_Y
    packet += recv_path[0] + recv_path[1] + copy_x + copy_y
    packet += f'{data_packet:030b}'
    # assign packet
    stimuli += packet + get_origin_dest_comment(path, recv_path) + "\n"
    # reference    
    """for i in range(int(len(recv_path)/2)):
        index_path = get_index_path(recv_path[2*i],recv_path[2*i+1])
        reference += "0 " +  str(index_path) + " "
        reference += packets[i] + "\n"
    """
    reference += get_multicast_reference(copy_x, copy_y, recv_path, packet)

    return data_packet, stimuli, reference


def generate_random_send_recv_paths(num_msgs, external = False):
    sender_paths = []
    recv_paths = []
    for i in range(num_msgs):
        origin_type = np.random.randint(4)
        if (origin_type == 0) and external: # comes from external
            dir_index = np.random.randint(4)
            direction = directions[dir_index]
            sender_paths.append(direction)
        else:
            # generate origin (it shouldn't repeat)
            x_path = f'{np.random.randint(32):05b}'
            y_path = f'{np.random.randint(32):05b}'
            while [x_path, y_path] in sender_paths:
                x_path = f'{np.random.randint(32):05b}'
                y_path = f'{np.random.randint(32):05b}'
            sender_paths.append([x_path, y_path])

        # decide if it is unicast or multicast or external
        msg_type = np.random.randint(5)
        if (msg_type == 0) and external:
            dir_index = np.random.randint(4)
            direction = directions[dir_index]
            recv_paths.append(direction)
        elif msg_type <= 2: # unicast
            y_path = f'{np.random.randint(32):05b}'
            x_path = f'{np.random.randint(32):05b}'
            recv_paths.append([x_path, y_path])
        else: # multicast
            num_rcv = np.random.randint(2,7)
            recv_path = []
            for i in range(num_rcv):
                recv_path.append(f'{np.random.randint(32):05b}')
                recv_path.append(f'{np.random.randint(32):05b}')
            recv_paths.append(recv_path)
        
    return sender_paths, recv_paths
